LINES
COLS
stdscr
curscr
SP
Mouse_status
COLORS
COLOR_PAIRS
TABSIZE
acs_map
ttytype
addch
addchnstr
addchstr
addnstr
addstr
attroff
attron
attrset
attr_get
attr_off
attr_on
attr_set
baudrate
beep
bkgd
bkgdset
border
box
can_change_color
cbreak
chgat
clearok
clear
clrtobot
clrtoeol
color_content
color_set
copywin
curs_set
def_prog_mode
def_shell_mode
delay_output
delch
deleteln
delscreen
delwin
derwin
doupdate
dupwin
echochar
echo
endwin
erasechar
erase
filter
flash
flushinp
getbkgd
getnstr
getstr
getwin
halfdelay
has_colors
has_ic
has_il
hline
idcok
idlok
immedok
inchnstr
inchstr
inch
init_color
init_pair
initscr
innstr
insch
insdelln
insertln
insnstr
insstr
instr
intrflush
isendwin
is_linetouched
is_wintouched
keyname
keypad
killchar
leaveok
longname
meta
move
mvaddch
mvaddchnstr
mvaddchstr
mvaddnstr
mvaddstr
mvchgat
mvcur
mvdelch
mvderwin
mvgetch
mvgetnstr
mvgetstr
mvhline
mvinch
mvinchnstr
mvinchstr
mvinnstr
mvinsch
mvinsnstr
mvinsstr
mvinstr
mvprintw
mvscanw
mvvline
mvwaddchnstr
mvwaddchstr
mvwaddch
mvwaddnstr
mvwaddstr
mvwchgat
mvwdelch
mvwgetch
mvwgetnstr
mvwgetstr
mvwhline
mvwinchnstr
mvwinchstr
mvwinch
mvwinnstr
mvwinsch
mvwinsnstr
mvwinsstr
mvwinstr
mvwin
mvwprintw
mvwscanw
mvwvline
napms
newpad
newterm
newwin
nl
nocbreak
nodelay
noecho
nonl
noqiflush
noraw
notimeout
overlay
overwrite
pair_content
pechochar
pnoutrefresh
prefresh
printw
putwin
qiflush
raw
redrawwin
refresh
reset_prog_mode
reset_shell_mode
resetty
ripoffline
savetty
scanw
scr_dump
scr_init
scrl
scrollok
scroll
scr_restore
scr_set
setscrreg
setsyx
set_term
slk_attroff
slk_attr_off
slk_attron
slk_attr_on
slk_attrset
slk_attr_set
slk_clear
slk_color
slk_init
slk_label
slk_noutrefresh
slk_refresh
slk_restore
slk_set
slk_touch
standend
standout
start_color
subpad
subwin
syncok
termattrs
term_attrs
termname
timeout
touchline
touchwin
typeahead
untouchwin
use_env
vidattr
vid_attr
vidputs
vid_puts
vline
vw_printw
vwprintw
vw_scanw
vwscanw
waddchnstr
waddchstr
waddch
waddnstr
waddstr
wattroff
wattron
wattrset
wattr_get
wattr_off
wattr_on
wattr_set
wbkgdset
wbkgd
wborder
wchgat
wclear
wclrtobot
wclrtoeol
wcolor_set
wcursyncup
wdelch
wdeleteln
wechochar
werase
wgetch
wgetnstr
wgetstr
whline
winchnstr
winchstr
winch
winnstr
winsch
winsdelln
winsertln
winsnstr
winsstr
winstr
wmove
wnoutrefresh
wprintw
wredrawln
wrefresh
wscanw
wscrl
wsetscrreg
wstandend
wstandout
wsyncdown
wsyncup
wtimeout
wtouchln
wvline
getattrs
getbegx
getbegy
getmaxx
getmaxy
getparx
getpary
getcurx
getcury
traceoff
traceon
unctrl
crmode
nocrmode
draino
resetterm
fixterm
saveterm
mouse_set
mouse_on
mouse_off
request_mouse_pos
map_button
wmouse_position
getmouse
getbmap
assume_default_colors
curses_version
has_key
use_default_colors
wresize
mouseinterval
mousemask
mouse_trafo
nc_getmouse
ungetmouse
wenclose
wmouse_trafo
addrawch
insrawch
is_termresized
mvaddrawch
mvdeleteln
mvinsertln
mvinsrawch
mvwaddrawch
mvwdeleteln
mvwinsertln
mvwinsrawch
raw_output
resize_term
resize_window
waddrawch
winsrawch
wordchar
bottom_panel
del_panel
hide_panel
move_panel
new_panel
panel_above
panel_below
panel_hidden
panel_userptr
panel_window
replace_panel
set_panel_userptr
show_panel
top_panel
update_panels
PDC_debug
PDC_ungetch
PDC_set_blink
PDC_set_line_color
PDC_set_title
PDC_clearclipboard
PDC_freeclipboard
PDC_getclipboard
PDC_setclipboard
PDC_get_input_fd
PDC_get_key_modifiers
PDC_save_key_modifiers
PDC_return_key_modifiers
cur_term
del_curterm
putp
restartterm
set_curterm
setterm
setupterm
tgetent
tgetflag
tgetnum
tgetstr
tgoto
tigetflag
tigetnum
tigetstr
tparm
tputs
