addnwstr
addwstr
add_wch
add_wchnstr
add_wchstr
border_set
box_set
echo_wchar
erasewchar
getbkgrnd
getcchar
getn_wstr
get_wch
get_wstr
hline_set
innwstr
ins_nwstr
ins_wch
ins_wstr
inwstr
in_wch
in_wchnstr
in_wchstr
key_name
killwchar
mvaddnwstr
mvaddwstr
mvadd_wch
mvadd_wchnstr
mvadd_wchstr
mvgetn_wstr
mvget_wch
mvget_wstr
mvhline_set
mvinnwstr
mvins_nwstr
mvins_wch
mvins_wstr
mvinwstr
mvin_wch
mvin_wchnstr
mvin_wchstr
mvvline_set
mvwaddnwstr
mvwaddwstr
mvwadd_wch
mvwadd_wchnstr
mvwadd_wchstr
mvwgetn_wstr
mvwget_wch
mvwget_wstr
mvwhline_set
mvwinnwstr
mvwins_nwstr
mvwins_wch
mvwins_wstr
mvwin_wch
mvwin_wchnstr
mvwin_wchstr
mvwinwstr
mvwvline_set
pecho_wchar
setcchar
slk_wset
unget_wch
vline_set
waddnwstr
waddwstr
wadd_wch
wadd_wchnstr
wadd_wchstr
wbkgrnd
wbkgrndset
wborder_set
wecho_wchar
wgetbkgrnd
wgetn_wstr
wget_wch
wget_wstr
whline_set
winnwstr
wins_nwstr
wins_wch
wins_wstr
winwstr
win_wch
win_wchnstr
win_wchstr
wunctrl
wvline_set
