#include "winver.h"

1 VERSIONINFO
 FILEVERSION 3,4,0,0
 PRODUCTVERSION 3,4,0,0
 FILEFLAGSMASK 0x3fL
 FILEFLAGS 0x0L
 FILEOS VOS_UNKNOWN
 FILETYPE VFT_DLL
 FILESUBTYPE 0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904b0"
        BEGIN
            VALUE "Maintainer", "William McBrine\0"
            VALUE "FileDescription", "Public Domain Curses\0"
            VALUE "FileVersion", "3.4.0\0"
            VALUE "InternalName", "PDCurses\0"
            VALUE "LegalCopyright", "Public Domain\0"
            VALUE "OriginalFilename", "PDCURSES.DLL\0"
            VALUE "ProductName", "Public Domain Curses Library\0"
            VALUE "ProductVersion", "3.4.0\0"
        END
    END
END

1 ICON    DISCARDABLE     "PDCURSES.ico"
